## ----setup, include = FALSE----------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library("GLMMadaptive")

## ------------------------------------------------------------------------
set.seed(1234)
n <- 100 # number of subjects
K <- 8 # number of measurements per subject
t_max <- 15 # maximum follow-up time

# we constuct a data frame with the design: 
# everyone has a baseline measurment, and then measurements at random follow-up times
DF <- data.frame(id = rep(seq_len(n), each = K),
                 time = c(replicate(n, c(0, sort(runif(K - 1, 0, t_max))))),
                 sex = rep(gl(2, n/2, labels = c("male", "female")), each = K))

# design matrices for the fixed and random effects
X <- model.matrix(~ sex * time, data = DF)
Z <- model.matrix(~ time, data = DF)

betas <- c(-2.13, -0.25, 0.24, -0.05) # fixed effects coefficients
D11 <- 0.48 # variance of random intercepts
D22 <- 0.1 # variance of random slopes

# we simulate random effects
b <- cbind(rnorm(n, sd = sqrt(D11)), rnorm(n, sd = sqrt(D22)))
# linear predictor
eta_y <- as.vector(X %*% betas + rowSums(Z * b[DF$id, ]))
# we simulate binary longitudinal data
DF$y <- rbinom(n * K, 1, plogis(eta_y))

## ------------------------------------------------------------------------
fm <- mixed_model(fixed = y ~ sex * time, random = ~ time | id, data = DF,
                  family = binomial())

## ------------------------------------------------------------------------
fm

## ------------------------------------------------------------------------
summary(fm)

## ------------------------------------------------------------------------
exp(confint(fm))

## ------------------------------------------------------------------------
confint(fm, parm = "var-cov", level = 0.90)

## ------------------------------------------------------------------------
vcov(fm)

## ------------------------------------------------------------------------
fixef(fm)

## ------------------------------------------------------------------------
head(ranef(fm))

## ------------------------------------------------------------------------
head(coef(fm))

## ------------------------------------------------------------------------
marginal_coefs(fm)

## ---- eval = FALSE-------------------------------------------------------
#  marginal_coefs(fm, std_errors = TRUE)

## ------------------------------------------------------------------------
head(fitted(fm))

## ------------------------------------------------------------------------
head(fitted(fm, type = "subject_specific"))

## ---- eval = FALSE-------------------------------------------------------
#  head(fitted(fm, type = "marginal"))

## ------------------------------------------------------------------------
nDF <- with(DF, expand.grid(time = seq(min(time), max(time), length.out = 15),
                            sex = levels(sex)))

## ------------------------------------------------------------------------
plot_data <- effectPlotData(fm, nDF)

## ---- fig.show="hold"----------------------------------------------------
library("lattice")
xyplot(pred + low + upp ~ time | sex, data = plot_data,
       type = "l", lty = c(1, 2, 2), col = c(2, 1, 1), lwd = 2,
       xlab = "Follow-up time", ylab = "log odds")

expit <- function (x) exp(x) / (1 + exp(x))
xyplot(expit(pred) + expit(low) + expit(upp) ~ time | sex, data = plot_data,
       type = "l", lty = c(1, 2, 2), col = c(2, 1, 1), lwd = 2,
       xlab = "Follow-up time", ylab = "Subject-Specific Probabilities")

## ---- eval = FALSE-------------------------------------------------------
#  plot_data_m <- effectPlotData(fm, nDF, marginal = TRUE)
#  
#  xyplot(expit(pred) + expit(low) + expit(upp) ~ time | sex, data = plot_data_m,
#         type = "l", lty = c(1, 2, 2), col = c(2, 1, 1), lwd = 2,
#         xlab = "Follow-up time", ylab = "Marginal Probabilities")

## ------------------------------------------------------------------------
gm <- mixed_model(fixed = y ~ sex * time, random = ~ time || id, data = DF,
                  family = binomial())

anova(gm, fm)

